;===============================================================================
; 28CXXX EEPROM Programmer
;===============================================================================
;
; Notes:
;
; Operations that write to the EEPROM must perform consecutive updates within
; 150uSec of each other to remain in page write mode.
;
;-------------------------------------------------------------------------------

#define M(X)    (.1<<(X))

;===============================================================================
; ASCII Constants
;-------------------------------------------------------------------------------

BS              equ     h'08'
LF              equ     h'0a'
CR              equ     h'0d'
DEL             equ     h'7f'
DC1             equ     h'11'
DC3             equ     h'13'

;===============================================================================
; Device Configuration
;-------------------------------------------------------------------------------

                ifdef   __18F4620
                include "P18F4620.inc"

                config  OSC=INTIO67,FCMEN=OFF,IESO=OFF
                config  CCP2MX=PORTC,PBADEN=OFF,LPT1OSC=OFF
                config  MCLRE=ON,STVREN=ON,LVP=OFF,XINST=OFF

                ifdef   __DEBUG
                config  PWRT=OFF,BOREN=OFF,WDT=OFF,DEBUG=ON
                config  CP0=OFF,CP1=OFF,CP2=OFF,CP3=OFF,CPB=OFF,CPD=OFF
                config  WRT0=OFF,WRT1=OFF,WRT2=OFF,WRT3=OFF,WRTC=OFF,WRTB=OFF,WRTD=OFF
                config  EBTR0=OFF,EBTR1=OFF,EBTR2=OFF,EBTR3=OFF,EBTRB=OFF
                else
                config  PWRT=ON,BOREN=OFF,WDT=ON,WDTPS=8192,DEBUG=OFF
                config  CP0=ON,CP1=ON,CP2=ON,CP3=ON,CPB=ON,CPD=ON
                config  WRT0=ON,WRT1=ON,WRT2=ON,WRT3=ON,WRTC=ON,WRTB=ON,WRTD=ON
                config  EBTR0=ON,EBTR1=ON,EBTR2=ON,EBTR3=ON,EBTRB=ON
                endif

OSC             equ     .8000000
PLL             equ     .4

ADDRL_TRIS      equ     TRISA
ADDRL_LAT       equ     LATA
ADDRH_TRIS      equ     TRISB
ADDRH_LAT       equ     LATB

DATA_TRIS       equ     TRISD
DATA_PORT       equ     PORTD
DATA_LAT        equ     LATD

CTRL_TRIS       equ     TRISE
CTRL_LAT        equ     LATE
RD_PIN          equ     .0
WR_PIN          equ     .1
CS_PIN          equ     .2

                endif

;===============================================================================
;-------------------------------------------------------------------------------

FOSC            equ     OSC * PLL

BRG_9600        equ     FOSC / (.16 * .9600) - .1

;===============================================================================
; Data Areas
;-------------------------------------------------------------------------------

                udata_acs

LENGTH          res     .1              ; Command buffer length

CHAR            res     .1              ; Character buffer
TEMP            res     .1              ; Scratch space

BASE            res     .2              ; Base address of the EEPROM in target
SIZE            res     .2              ; The size of the EEPROM

BYTE            res     .1
WORD            res     .2

START           res     .2              ; Start address for command
FINAL           res     .2              ; Final address for command

NUMBYTES        res     .1              ; Number of bytes in S19 record
CHECKSUM        res     .1              ; The checksum

BYTECOUNT       res     .1              ; Number of bytes to write

LAST_BYTE       res     .1              ; The last byte written

                udata

BINARY          res     .64             ; Buffer for binary data
COMMAND         res     .128            ; Command line buffer

;===============================================================================
; Power On Reset
;-------------------------------------------------------------------------------

.ResetVector    code    h'0000'

                goto    PowerOnReset

;-------------------------------------------------------------------------------

                code
PowerOnReset:
                movlw   b'00001111'     ; Make all pins digital
                movwf   ADCON1
                movlw   b'00000111'     ; And comparators turned off
                movwf   CMCON

                clrf    CTRL_TRIS       ; Make CS/RD/WR HI outputs
                setf    CTRL_LAT
                clrf    ADDRL_TRIS      ; Make ADDR pins LO outputs
                clrf    ADDRH_TRIS
                clrf    ADDRL_LAT
                clrf    ADDRH_LAT
                clrf    DATA_TRIS       ; Make DATA pins LO outputs
                clrf    DATA_PORT

                clrf    TRISC           ; Make unused pins LO
                clrf    LATC

;-------------------------------------------------------------------------------

                movlw   M(PLLEN)        ; Enable PLL module
                movwf   OSCTUNE
                movlw   b'01110000'     ; Select 8Mhz INT osc
                movwf   OSCCON

                ifndef  __DEBUG
WaitTillStable:
                btfss   OSCCON,IOFS
                bra     WaitTillStable
                endif

;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------

                bsf     TRISC,.7        ; Make RXD pin an input
                bcf     TRISC,.6        ; Make TXD pin an output

                movlw   M(BRG16)        ; Set baud rate
                movwf   BAUDCON
                movlw   low BRG_9600
                movwf   SPBRG
                movlw   high BRG_9600
                movwf   SPBRGH

                movlw   M(TXEN)|M(BRGH) ; Enable transmit
                movwf   TXSTA
                movlw   M(SPEN)|M(CREN) ; Enable receive
                movwf   RCSTA

                movf    RCREG,W         ; Clear input FIFO
                movf    RCREG,W

;-------------------------------------------------------------------------------

                rcall   NewLine
                rcall   NewLine
                rcall   Print
                db      "EP-28CXXX (C),2013 HandCoded Software Ltd.",CR,LF,.0
                rcall   NewLine
                rcall   Print
                db      "Type '?' for a list of commands",CR,LF,.0
                rcall   NewLine         

;-------------------------------------------------------------------------------

                clrf    BASE+.0
                clrf    BASE+.1
                movlw   low  (.16 * .1024)
                movwf   SIZE+.0
                movlw   high (.16 * .1024)
                movwf   SIZE+.1

;===============================================================================
; Command Input
;-------------------------------------------------------------------------------

ReadCommand:
                clrf    STKPTR          ; Reset the stack pointer
                movlw   '>'             ; Display the command prompt
                rcall   UartTx

                lfsr    .0,COMMAND      ; Point FSR0 at the command buffer
                clrf    LENGTH          ; .. and reset the length

                movlw   DC1             ; Send XON
                rcall   UartTx

ReadCharacter:
                rcall   UartRx          ; Read and save the next character
                movwf   CHAR

                xorlw   DEL             ; Check for control characters
                bz      DoDelete
                xorlw   CR^DEL
                bz      EndOfInput

                movf    CHAR,W          ; Ignore all other characters
                rcall   IsCtrl
                bc      ReadCharacter

                movf    LENGTH,W        ; Save character at end of buffer
                movff   CHAR,PLUSW0
                incf    LENGTH,F
                movf    CHAR,W          ; Echo character to the diaplay
                rcall   UartTx
                bra     ReadCharacter   ; And repeat

;-------------------------------------------------------------------------------

DoDelete:
                movf    LENGTH,W        ; Any characters in the buffer? 
                bz      ReadCharacter   ; No.

                movlw   BS              ; Erase the last character
                rcall   UartTx
                movlw   ' '
                rcall   UartTx
                movlw   BS
                rcall   UartTx

                decf    LENGTH,F        ; Decrease the character count
                bra     ReadCharacter

;-------------------------------------------------------------------------------

EndOfInput:
                movlw   DC3             ; Send XOFF
                rcall   UartTx

                rcall   NewLine
                rcall   SkipSpaces
                bz      ReadCommand
                rcall   ToUpper
                
;-------------------------------------------------------------------------------

SetBase:
                xorlw   'B'
                bnz     FillMemory     

                rcall   SkipSpaces      ; Skip to start address
                rcall   ReadWord        ; And try to read it
                movff   WORD+.0,BASE+.0 ; Copy to BASE address
                movff   WORD+.1,BASE+.1

                rcall   Print           ; Echo the new base setting
                db      "BASE=",0
                rcall   ToHexWord
                rcall   NewLine

                goto    ReadCommand     ; And read another command

;-------------------------------------------------------------------------------

FillMemory:
                xorlw   'F'^'B'
                bnz     ShowMemory

                rcall   SkipSpaces      ; Skip to start address
                rcall   ReadWord        ; Try to read it
                movff   WORD+.0,START+.0
                movff   WORD+.1,START+.1
                rcall   SkipSpaces      ; Skip to final address
                rcall   ReadWord
                movff   WORD+.0,FINAL+.0
                movff   WORD+.1,FINAL+.1
                rcall   SkipSpaces
                rcall   ReadByte
                movwf   CHAR

                movf    BASE+.0,W       ; Calculate offset into EEPROM
                subwf   START+.0,W
                movwf   ADDRL_LAT
                movf    BASE+.1,W
                subwfb  START+.1,W
                movwf   ADDRH_LAT
FillLoop:
                movf    CHAR,W          ; Fill the target address
                rcall   ChipWrite
                infsnz  ADDRL_LAT,F     ; Then bump the address
                incf    ADDRH_LAT,F
                
                movlw   .1              ; Work out the next address
                addwf   START+.0,F
                movlw   .0
                addwfc  START+.1,F
                bc      FillDone        ; Stop if wrapped round

                movf    START+.0,W      ; Compare with end address
                subwf   FINAL+.0,W
                movf    START+.1,W
                subwfb  FINAL+.1,W
                bc      FillLoop        ; Not reached the end yet
FillDone:
                rcall   ChipWait
                goto    ReadCommand     ; And read another command

;-------------------------------------------------------------------------------

ShowMemory:
                xorlw   'M'^'F'
                bnz     LoadS19

                rcall   SkipSpaces      ; Skip to start address
                rcall   ReadWord        ; Try to read it
                movff   WORD+.0,START+.0
                movff   WORD+.1,START+.1
                rcall   SkipSpaces      ; Skip to final address
                rcall   ReadWord
                movff   WORD+.0,FINAL+.0
                movff   WORD+.1,FINAL+.1

ReadLoop:
                movff   START+.0,WORD+.0 ; Show start address
                movff   START+.1,WORD+.1
                rcall   ToHexWord

                movf    BASE+.0,W       ; Calculate offset into EEPROM
                subwf   WORD+.0,F
                movf    BASE+.1,W
                subwfb  WORD+.1,F

                movff   WORD+.0,ADDRL_LAT ; Set target address
                movff   WORD+.1,ADDRH_LAT
                movlw   .16             ; Set byte counter
                movwf   LENGTH

ReadByteLoop:
                movlw   ' '
                rcall   UartTx
                rcall   ChipRead
                rcall   ToHexByte
                
                infsnz  ADDRL_LAT,F     ; Bump memory address
                incf    ADDRH_LAT,F                            

                decfsz  LENGTH,F
                bra     ReadByteLoop

                movff   WORD+.0,ADDRL_LAT ; Set target address
                movff   WORD+.1,ADDRH_LAT
                movlw   .16             ; Set byte counter
                movwf   LENGTH

                movlw   ' '
                rcall   UartTx
                movlw   ':'
                rcall   UartTx
                movlw   ' '
                rcall   UartTx

ReadCharLoop:
                rcall   ChipRead        ; Read from the chip

                andlw   h'7f'           ; Make printable
                rcall   IsCtrl
                btfsc   STATUS,C
                movlw   '.'
                rcall   UartTx

                infsnz  ADDRL_LAT,F     ; Bump memory address
                incf    ADDRH_LAT,F                            

                decfsz  LENGTH,F
                bra     ReadCharLoop

                rcall   NewLine

                movlw   .16             ; Work out the next address
                addwf   START+.0,F
                movlw   .0
                addwfc  START+.1,F
                bc      ShowDone        ; Stop if wrapped round

                movf    START+.0,W      ; Compare with end address
                subwf   FINAL+.0,W
                movf    START+.1,W
                subwfb  FINAL+.1,W
                bc      ReadLoop        ; Not reached the end yet
ShowDone:
                goto    ReadCommand     ; And read another command

;-------------------------------------------------------------------------------

LoadS19:
                xorlw   'S'^'M'
                bnz     ShowHelp

                rcall   NextChar        ; Get the record type
                xorlw   '1'             ; Only process data records
                bz      LoadRecord
                goto    ReadCommand

LoadRecord:
                rcall   NextChar
                rcall   ReadByte        ; Fetch the data length
                movwf   NUMBYTES        ; .. and save
                movwf   CHECKSUM

                rcall   NextChar
                rcall   ReadWord        ; Fetch the target address
                decf    NUMBYTES,F
                bz      InvalidS19
                decf    NUMBYTES,F
                bz      InvalidS19

;        movlw   '-'
;        rcall   UartTx
;        rcall   ToHexWord

                movf    WORD+.0,W       ; Add address into checksum
                addwf   CHECKSUM,F
                movf    WORD+.1,W
                addwf   CHECKSUM,F


;        movlw ' '
;        rcall   UartTx
;        movf    CHECKSUM,W
;        rcall   ToHexByte
;        movlw   ' '
;        rcall   UartTx

                lfsr    .1,BINARY       ; Point FSR1 at data buffer
                clrf    BYTECOUNT       ; And clear the count

LoadByte:
                rcall   NextChar        ; Fetch a data byte
                rcall   ReadByte

                decf    NUMBYTES,F      ; Reached the last byte?
                bz      LoadDone
                addwf   CHECKSUM,F

                movwf   POSTINC1        ; Save data byte in buffer
                incf    BYTECOUNT,F     ; And bump counter

;        movf    DATA_LAT,W
;        rcall   ToHexByte

                bra     LoadByte                         

LoadDone:
                movwf   NUMBYTES
                comf    CHECKSUM,F
;        movlw '='
;        rcall   UartTx
;        movf    CHECKSUM,W
;        rcall   ToHexByte
;        movlw '|'
;        rcall   UartTx
;        movf    NUMBYTES,W
;        rcall   ToHexByte
;        rcall   NewLine

                movf    NUMBYTES,W
                xorwf   CHECKSUM,W      ; Correct checksum?
                bnz     InvalidChecksum

                movf    BASE+.0,W       ; Calculate offset into EEPROM
                subwf   WORD+.0,W       ; .. and set target address
                movwf   ADDRL_LAT
                movf    BASE+.1,W
                subwfb  WORD+.1,W
                movwf   ADDRH_LAT

                movf    BYTECOUNT,F     ; Any bytes to write?
                bz      WriteDone       ; No.
                lfsr    .1,BINARY       ; Point FSR1 at data buffer
WriteLoop:
                movf    POSTINC1,W      ; Fetch a buffered data byte
                rcall   ChipWrite       ; And write to EEPROM

                decf    BYTECOUNT,F     ; More bytes left to write?
                bz      WriteWait

                infsnz  ADDRL_LAT,F     ; Bump memory address
                incf    ADDRH_LAT,F   
                bra     WriteLoop

WriteWait:
                rcall   ChipWait        ; Wait for last write to finish
WriteDone:
                goto    ReadCommand     ; Yes, get another line

InvalidS19:
                rcall   Print
                db      "Invalid S19 record",CR,LF,.0
                goto    ReadCommand

InvalidChecksum:
                rcall   Print
                db      "Invalid checksum",CR,LF,.0
                goto    ReadCommand

;-------------------------------------------------------------------------------

ShowHelp:
                xorlw   '?'^'S'
                bnz     UnknownCommand

                rcall   Print
                db      " B xxxx\t\t\tSet EEPROM base address",CR,LF,.0
                rcall   Print
                db      " F xxxx yyyy bb\t\tFill memory from xxxx to yyyy with bb",CR,LF,.0
                rcall   Print
                db      " M xxxx [yyyy]\t\tDump memory from xxxx to yyyy",CR,LF,.0
                rcall   Print
                db      " Sxxxxxxxxxxxx\t\tLoad an S19 data record",CR,LF,.0
                rcall   Print
                db      " ?\t\t\tShow Help",CR,LF,.0
                goto    ReadCommand

;-------------------------------------------------------------------------------

UnknownCommand:
                rcall   Print
                db      "Unknown command",CR,LF,.0
                goto    ReadCommand

;-------------------------------------------------------------------------------

ReadWord:
                rcall   ReadByte
                movwf   WORD+.1
                rcall   NextChar
                rcall   ReadByte
                movwf   WORD+.0
                return

ReadByte:
                rcall   IsHexDigit      ; Got a hex digit?
                bnc     InvalidHex      ; No.
                rcall   ToUpper         ; Make upper case
                btfsc   WREG,.6         ; Is it a digit?
                addlw   -.7             ; Convert to hex
                andlw   h'0f'
                swapf   WREG
                movwf   BYTE

                rcall   NextChar
                rcall   IsHexDigit      ; Got a hex digit?
                bnc     InvalidHex      ; No.
                rcall   ToUpper         ; Make upper case
                btfsc   WREG,.6         ; Is it a digit?
                addlw   -.7             ; Convert to hex
                andlw   h'0f'
                iorwf   BYTE,W          ; And combine
                return

;

InvalidHex:
                rcall   Print
                db      "Invalid HEX",CR,LF,.0
                goto    ReadCommand

;-------------------------------------------------------------------------------

; Return the next character from the input buffer. If we run out of data then
; return zero bytes instead.

NextChar:
                movf    LENGTH,W        ; Reached the end of buffer?
                btfsc   STATUS,Z
                return                  ; Yes
                decf    LENGTH,F        ; No, extract a character
                movf    POSTINC0,W
                return
                
; Read the characters from buffer until we reach a non-space.

SkipSpaces:
                rcall   NextChar        ; Extract a character
                xorlw   ' '             ; Is it a space?
                bz      SkipSpaces      ; Yes
                xorlw   ' '             ; Recover the character
                return                  ; Done

;===============================================================================
;-------------------------------------------------------------------------------

;

IsLower:
                movwf   TEMP            ; Save the character
                addlw   -'a'            ; Compare with 'a'
                bnc     Fail            ; Less?
                movf    TEMP,W        
                addlw   -('z'+.1)       ; Compare with 'z'
                bc      Fail            ; Greater?
                bra     Pass

;

IsHexDigit:
                rcall   IsDigit         ; Got a digit?
                btfsc   STATUS,C
                return                  ; Yes.

                movwf   TEMP            ; Save the character
                addlw   -'A'            ; A-F?
                bnc     Fail            ; No.
                movf    TEMP,W
                addlw   -('F'+.1)
                bnc     Pass            ; Yes.
                movf    TEMP,W
                addlw   -'a'            ; a-f?
                bnc     Fail            ; No.
                movf    TEMP,W
                addlw   -('f'+.1)
                bnc     Pass            ; Yes.
                bra     Fail

;

IsDigit:
                movwf   TEMP            ; Save the character
                addlw   -'0'            ; Compare with '0'
                bnc     Fail            ; Less?
                movf    TEMP,W
                addlw   -('9'+.1)       ; Compare with '9'
                bc      Fail            ; Greater?
                bra     Pass
;

IsCtrl:
                movwf   TEMP            ; Save the characater
                xorlw   DEL             ; DEL character?
                bz      Pass
                movf    TEMP,W  
                addlw   -h'20'          ; 00-1f?
                bnc     Pass
                bra     Fail

Pass:
                bsf     STATUS,C        ; Set carry to indicate success
                movf    TEMP,W          ; Restore character
                return

Fail:
                bcf     STATUS,C        ; Clear carry to indicate failure
                movf    TEMP,W          ; Restore character
                return

;-------------------------------------------------------------------------------

; Convert the character in WREH to UPPER case.

ToUpper:
                rcall   IsLower         ; Got a lower character character?
                btfsc   STATUS,C
                addlw   -.32            ; Convert to upper case
                return

;

ToHexWord:
                movf    WORD+.1,W
                rcall   ToHexByte
                movf    WORD+.0,W

ToHexByte:
                movwf   TEMP
                swapf   WREG,W
                rcall   ToHexDigit
                rcall   UartTx
                movf    TEMP,W
                rcall   ToHexDigit
                bra     UartTx

ToHexDigit:
                andlw   h'0f'
                mullw   .2
                rcall   ComputedJump

                retlw   '0'
                retlw   '1'
                retlw   '2'
                retlw   '3'
                retlw   '4'
                retlw   '5'
                retlw   '6'
                retlw   '7'
                retlw   '8'
                retlw   '9'
                retlw   'A'
                retlw   'B'
                retlw   'C'
                retlw   'D'
                retlw   'E'
                retlw   'F'

;-------------------------------------------------------------------------------

ComputedJump:
                movf    PRODL,W
                addwf   TOSL,F
                movf    PRODH,W
                addwfc  TOSH,F
                return

;===============================================================================
; EEPROM Routines
;-------------------------------------------------------------------------------

ChipRead:
                setf    DATA_TRIS       ; Set data port to read

                bcf     CTRL_LAT,CS_PIN ; Select the chip
                bcf     CTRL_LAT,RD_PIN ; Enable read
                nop
                nop
                movf    DATA_PORT,W     ; Read a data byte
                bsf     CTRL_LAT,RD_PIN
                bsf     CTRL_LAT,CS_PIN
                return

;-------------------------------------------------------------------------------

ChipWrite:
                clrf    DATA_TRIS       ; Set data port to write

                movwf   DATA_LAT        ; Save the new value
                movwf   LAST_BYTE
                nop
                nop

                bcf     CTRL_LAT,CS_PIN ; Select the chip
                bcf     CTRL_LAT,WR_PIN ; Enable read
                nop
                nop
                bsf     CTRL_LAT,WR_PIN
                bsf     CTRL_LAT,CS_PIN

                movf    ADDRL_LAT,W     ; Reached end of page?
                andlw   .63
                xorlw   .63
                bz      ChipWait
                return
             
ChipWait:  
                setf    DATA_TRIS       ; Set data port to read

                nop
                nop
                nop
                nop

                bcf     CTRL_LAT,CS_PIN ; Select the chip
                bcf     CTRL_LAT,RD_PIN ; Enable read
                nop
                nop
                movf    DATA_PORT,W     ; Read a data byte
                bsf     CTRL_LAT,RD_PIN
                bsf     CTRL_LAT,WR_PIN

                xorwf   LAST_BYTE,W     ; Has the write finished?
                andlw   h'80'
                bnz     ChipWait
                return

;===============================================================================
; UART Interface
;-------------------------------------------------------------------------------

UartTx:
                btfss   PIR1,TXIF       ; Wait for space in TX FIFO
                bra     UartTx
                movwf   TXREG
                return

UartRx:
                btfss   PIR1,RCIF       ; Wait for data in RX FIFO
                bra     UartRx
                movf    RCREG,W
                return

;-------------------------------------------------------------------------------

NewLine:
                movlw   CR              ; Output a CR/LF sequence
                rcall   UartTx
                movlw   LF
                bra     UartTx

Print:
                movf    TOSL,W          ; Copy return address to table pointer
                movwf   TBLPTRL
                movf    TOSH,W
                movwf   TBLPTRH
                movf    TOSU,W
                movwf   TBLPTRU

PrintLoop:
                tblrd   *+              ; Fetch a character
                movf    TABLAT,W        ; End of string?
                bz      PrintDone
                rcall   UartTx          ; No, print it
                bra     PrintLoop       ; And repeat                

PrintDone:
                btfsc   TBLPTRL,.0      ; Finished on even address?
                tblrd   *+              ; No, bump address by one

                movf    TBLPTRL,W       ; Copy return address back to stack
                movwf   TOSL
                movf    TBLPTRH,W
                movwf   TOSH
                return                  ; Done

                end